# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

from st_library.utils.api_client.services.unstructured_data import UnstructuredDataService


# import of Query is at end of module as we have a circular dependency of
# Query.execute().results -> Table and Table.sample() -> Query

class Item(object):
    """
    Represents an unstructured Item object

    Parameters
    ----------
    matrix_id
    dataset_id

    """

    # Allowed characters in a BigQuery table column name
    _VALID_COLUMN_NAME_CHARACTERS = '_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

    # When fetching table contents, the max number of rows to fetch per HTTP request
    _DEFAULT_PAGE_SIZE = 1024

    # Milliseconds per week
    _MSEC_PER_WEEK = 7 * 24 * 3600 * 1000

    def __init__(self, matrix_id=None, dataset_id=None):
        self._service = UnstructuredDataService()
        self._datasets_id = dataset_id
        self._cached_page = None
        self._cached_page_index = 0
        self._schema = None

    def download_file(self, filename=None):
        """
        Download dataset files to disk.

        Parameters
        ----------
        filename
            the name of the dataset files.

        """
        raw_data = self._service.download_object(self._datasets_id, filename)

        with open(filename, 'wb') as f:
            f.write(raw_data)

        return filename

    def upload_file(self, filename=None, filepath=None):
        """
        Upload files to gcs.

        Parameters
        ----------
        filename
            the name of the file.
        filepath
            the path to the file

        """
        return self._service.upload_object(self._datasets_id, filename, filepath)

    def delete_file(self, filename=None):
        """
        Delete files from gcs.

        Parameters
        ----------
        filename
            the name of the file.

        """
        return self._service.delete_object(self._datasets_id, filename)



class Singleton(object):
    _instances = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            # noinspection PyArgumentList
            cls._instances[cls] = super(Singleton, cls).__new__(cls, *args, **kwargs)
        return cls._instances[cls]



import os


class Store(object):
    def __init__(self):
        raise TypeError('This class should not be initialized')

    config_id = os.environ.get('CONFIGURATION_ID')

    if os.environ.get('SEC_REFRESH_TOKEN'):
        token = os.environ['SEC_REFRESH_TOKEN']
    else:
        token = None

    performance_id = os.environ.get('PERFORMANCE_ID')

    token_data = dict()

    host = os.environ.get('HOST', 'https://shortesttrack.com')



from collections import defaultdict

from st_library import exceptions
from st_library.utils.api_client import ApiClient
from st_library.utils.api_client.models import ScriptExecutionConfiguration
from st_library.utils.helpers.store import Store


class Metadata(object):
    """
    A Category for work with metadata.

    """

    def __init__(self):
        self._script_execution_configurations = defaultdict(ScriptExecutionConfiguration)

    def get_parameter_value(self, parameter_id):
        """
        Get parameter value from the Script Execution Configuration

        Parameters
        ----------
        parameter_id : str

        Returns
        -------
        object
            The parameter value

        """
        return self._script_execution_configurations[Store.config_id].get_parameter_value(parameter_id)

    def set_performance_output_parameter_value(self, parameter_id, parameter_value):
        """
        Set output parameter to the Performance

        Parameters
        ----------
        parameter_id : str
        parameter_value : object

        Raises
        ------
        st_library.exceptions.PerformanceRequired
            If Performance is not defined for current Environment

        """
        if not Store.performance_id:
            raise exceptions.PerformanceRequired('Performance is not defined for current Environment')

        performance_id = Store.performance_id
        ApiClient.post(ApiClient.res.performance_output_parameter_post(parameter_id, performance_id),
                       json={'value': parameter_value})

    @property
    def script_execution_configurations(self):
        return self._script_execution_configurations


_METADATA = Metadata()



from urlobject import URLObject
import os


ROOT = URLObject(os.environ.get('HOST', 'https://shortesttrack.com'))

UAA_SERVICE_ENDPOINT = ROOT.add_path('api/uaa')
OAUTH_SERVICE_ENDPOINT = ROOT.add_path('oauth')
JWTKEY_ENDPOINT = ROOT.add_path('oauth_jwtkey')
METADATA_SERVICE_ENDPOINT = ROOT.add_path('api/metadata')



from collections import OrderedDict

from st_library.exceptions import DatabaseConnectionError, DatabaseNotFound
from st_library.core.metadata.metadata import _METADATA


class ConnectorContainer(object):
    def __init__(self, *args, **kwargs):
        self._data_initialized = False
        self._list = list()
        self._dict = OrderedDict()

    def __getitem__(self, item):
        if isinstance(item, int):
            try:
                result = self._list[item]
            except IndexError:
                if self._data_initialized:
                    raise DatabaseNotFound('The database with index {} not found'.format(item))
                else:
                    self._initialize_data()
                    result = self._list[item]

            return result

        if isinstance(item, str):
            try:
                result = self._dict[item]
            except KeyError:
                if self._data_initialized:
                    raise DatabaseNotFound('The database with name "{}" not found'.format(item))
                else:
                    self._initialize_data()
                    result = self._dict[item]

            return result

        raise ValueError('Invalid key')

    def _initialize_data(self):
        try:
            self._do_initialize_data()
        except Exception as e:
            raise DatabaseConnectionError(str(e))

        self._data_initialized = True

    def _do_initialize_data(self):
        raise NotImplementedError

    @property
    def names(self):
        if not self._data_initialized:
            self._initialize_data()
        return list(self._dict.keys())

    def __repr__(self):
        names = self.names
        return repr(self._list)

    @property
    def default(self):
        names = self.names
        return self._list[0]

    def _fetch_param_dict(self, param_ids):
        param_dict = dict()
        for param_id in param_ids:
            param_dict[param_id] = _METADATA.get_parameter_value(param_id)
        return param_dict



import six

from st_library.utils.api_client import ApiClient
from st_library.utils.helpers.store import Store

MODE_PRINT = 'mode_print'
MODE_SEND = 'mode_send'

LEVEL_DEBUG = 'DEBUG'
LEVEL_INFO = 'INFO'
LEVEL_NOTICE = 'NOTICE'
LEVEL_WARNING = 'WARNING'
LEVEL_ERROR = 'ERROR'
LEVEL_CRITICAL = 'CRITICAL'
LEVEL_ALERT = 'ALERT'
LEVEL_EMERGENCY = 'EMERGENCY'


class Logger(object):
    """
    A Category for log messages

    Methods
    -------
    debug(message='foo', analysis=False)
        Log with DEBUG level

    info(message='foo', analysis=False)
        Log with INFO level

    notice(message='foo', analysis=False)
        Log with NOTICE level

    warning(message='foo', analysis=False)
        Log with WARNING level

    error(message='foo', analysis=False)
        Log with ERROR level

    critical(message='foo', analysis=False)
        Log with CRITICAL level

    alert(message='foo', analysis=False)
        Log with ALERT level

    emergency(message='foo', analysis=False)
        Log with EMERGENCY level

    Notes
    -----

    The target for logging data is depends on the existence of the performance_id for current Environment.
    If performance_id is specified, then all the data will be sent to ST Services.
    Otherwise, it will be just printed to stdout

    Examples
    --------

        >>> from st_library import Library
        >>> st_lib = Library()
        >>> st_lib.logger.info('Script started.')

    """
    def __init__(self):
        self._mode = MODE_SEND if Store.performance_id else MODE_PRINT

    def _format_message(self, message, level):
        return '{level}: {message}'.format(level=level, message=message)

    def _log(self, message, analysis, level):
        if self._mode == MODE_SEND:
            payload = {
                'level': level,
                'analysis': analysis,
                'data': message
            }
            ApiClient.post(ApiClient.res.logging_post(Store.performance_id), json=payload)
        else:
            six.print_(self._format_message(message, level))

    def debug(self, message, analysis=False):
        self._log(message, analysis, LEVEL_DEBUG)

    def info(self, message, analysis=False):
        self._log(message, analysis, LEVEL_INFO)

    def notice(self, message, analysis=False):
        self._log(message, analysis, LEVEL_NOTICE)

    def warning(self, message, analysis=False):
        self._log(message, analysis, LEVEL_WARNING)

    def error(self, message, analysis=False):
        self._log(message, analysis, LEVEL_ERROR)

    def critical(self, message, analysis=False):
        self._log(message, analysis, LEVEL_CRITICAL)

    def alert(self, message, analysis=False):
        self._log(message, analysis, LEVEL_ALERT)

    def emergency(self, message, analysis=False):
        self._log(message, analysis, LEVEL_EMERGENCY)



# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements HTTP client helper functionality."""
import datetime

import requests

from st_library import exceptions
from st_library.utils.helpers.store import Store
from st_library.utils.shared_toolkit.api_client import BaseApiClient, SecAccessTokenMixin

METHOD_GET = 'get'
METHOD_POST = 'post'
METHOD_DELETE = 'delete'


class DisabledLogger(object):
    def __getattr__(self, item):
        def _mock(*args, **kwargs):
            pass

        return _mock


_disabled_logger = DisabledLogger()


class _SharedApiClientAdapter(BaseApiClient, SecAccessTokenMixin):
    DELETE = 'delete'

    def _get_cache_sec_access_token_expiration(self, configuration_id):
        return Store.token_data.get('expiration')

    def _get_auth_string(self):
        return ''

    def get_sec_refresh_token(self, configuration_id):
        return Store.token

    def _set_cache_public_key(self, public_key):
        Store.token_data['public_key'] = public_key

    @property
    def _basic_auth_tuple(self):
        return ('script', 'noMatter',)

    def _raise_performing_request_error(self, *args, **kwargs):
        raise exceptions.ServiceRequestError(*args, **kwargs)

    def _set_cache_sec_access_token(self, configuration_id, access_token):
        Store.token_data['access_token'] = access_token

    def _raise_http_error(self, *args, **kwargs):
        raise exceptions.ServiceRequestError(*args, **kwargs)

    def _get_cache_public_key(self):
        return Store.token_data.get('public_key')

    def _get_cache_sec_access_token(self, configuration_id):
        return Store.token_data.get('access_token')

    def _set_cache_sec_access_token_expiration(self, configuration_id, expiration):
        Store.token_data['expiration'] = expiration

    def _get_logger(self):
        return _disabled_logger

    def request(self, *args, **kwargs):
        if kwargs.get('extra_headers') is None:
            kwargs['extra_headers'] = {}
        kwargs['extra_headers']['User-Agent'] = 'st-dataprovider/1.0'
        return super(_SharedApiClientAdapter, self).request(*args, **kwargs)


_adapter = _SharedApiClientAdapter()


def get(url, params=None, data=None, json=None, extra_headers=None, raw_response=False, stats=None):
    return _adapter.request(_adapter.GET, '', params=params, data=data, json=json, base=url,
                            extra_headers=extra_headers, sec_id_for_special_token=Store.config_id,
                            raw_content=raw_response)


def post(url, params=None, data=None, json=None, extra_headers=None, raw_response=False, stats=None):
    if data is None and json is None:
        raise exceptions.InternalError('Cannot perform POST request without data and json')

    return _adapter.request(_adapter.POST, '', params=params, data=data, json=json, base=url,
                            extra_headers=extra_headers, sec_id_for_special_token=Store.config_id,
                            raw_content=raw_response)


def delete(url, params=None, data=None, json=None, extra_headers=None, raw_response=False, stats=None):
    return _adapter.request(_adapter.DELETE, '', params=params, data=data, json=json, base=url,
                            extra_headers=extra_headers, sec_id_for_special_token=Store.config_id,
                            raw_content=raw_response)



import abc

import six

from . import resources


@six.add_metaclass(abc.ABCMeta)
class SecRefreshTokenMixin(object):
    @abc.abstractmethod
    def _get_cache_sec_refresh_token(self, configuration_id):
        """
        Return sec refresh token from cache

        """

    @abc.abstractmethod
    def _set_cache_sec_refresh_token(self, configuration_id, refresh_token):
        """
        Set sec refresh token to cache

        """

    def get_sec_refresh_token(self, configuration_id):
        refresh_token = self._get_cache_sec_refresh_token(configuration_id)

        if not refresh_token:
            json = self.request(self.GET, 'api-keys/{}'.format(configuration_id), base=resources.UAA_SERVICE_ENDPOINT)
            refresh_token = json['refreshTokenValue']
            self._set_cache_sec_refresh_token(configuration_id, refresh_token)

        return refresh_token



# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Table and View Schema APIs."""
import datetime
import pandas
from st_library import exceptions


class Schema(list):
    """
    Represents the schema of a BigQuery table as a flattened list of objects representing fields.
    Initializes from its raw JSON representation, a Pandas Dataframe, or a list.

    Parameters
    ----------
    definition
        a definition of the schema as a list of dictionaries with 'name' and 'type'
        entries and possibly 'mode' and 'description' entries. Only used if no data argument was
        provided. 'mode' can be 'NULLABLE', 'REQUIRED' or 'REPEATED'. For the allowed types, see:
        https://cloud.google.com/bigquery/preparing-data-for-bigquery#datatypes

    Notes
    -----

    Each field object has name, data_type, mode and description properties. Nested fields
    get flattened with their full-qualified names. So a Schema that has an object A with nested
    field B will be represented as [(name: 'A', ...), (name: 'A.b', ...)].

    """

    def __init__(self, definition=None):
        super(Schema, self).__init__()
        self._map = {}
        self._bq_schema = definition
        self._populate_fields(definition)

    class Field(object):
        """
        Represents a single field in a Table schema.

        Parameters
        ----------
        name: str
            the flattened, full-qualified name of the field.
        data_type: str
            the type of the field as a string ('INTEGER', 'BOOLEAN', 'FLOAT', 'STRING' or 'TIMESTAMP').
        mode: str
            the mode of the field; 'NULLABLE' by default.
        description: str
            a description of the field, if known; empty string by default.

        """

        # TODO(gram): consider renaming data_type member to type. Yes, it shadows top-level
        # name but that is what we are using in __str__ and __getitem__ and is what is used in BQ.
        # The shadowing is unlikely to cause problems.
        def __init__(self, name, data_type, mode='NULLABLE', description=''):
            self.name = name
            self.data_type = data_type
            self.mode = mode
            self.description = description

        def _repr_sql_(self):
            """

            Returns a representation of the field for embedding into a SQL statement.

            Returns
            -------
            A formatted field name for use within SQL statements.
            """
            return self.name

        def __eq__(self, other):
            """ Compare two schema field objects for equality (ignoring description). """
            return self.name == other.name and self.data_type == other.data_type \
                   and self.mode == other.mode

        def __str__(self):
            """ Returns the schema field as a string form of a dictionary. """
            return "{ 'name': '%s', 'type': '%s', 'mode':'%s', 'description': '%s' }" % \
                   (self.name, self.data_type, self.mode, self.description)

        def __repr__(self):
            """ Returns the schema field as a string form of a dictionary. """
            return str(self)

        def __getitem__(self, item):
            # TODO(gram): Currently we need this for a Schema object to work with the Parser object.
            # Eventually if we change Parser to only work with Schema (and not also with the
            # schema dictionaries in query results) we can remove this.

            if item == 'name':
                return self.name
            if item == 'type':
                return self.data_type
            if item == 'mode':
                return self.mode
            if item == 'description':
                return self.description

    @staticmethod
    def _from_dataframe(dataframe, default_type='STRING'):
        """
        Infer a BigQuery table schema from a Pandas dataframe.

        Parameters
        ----------
        dataframe
            The DataFrame.

        default_type
            The default big query type in case the type of the column does not exist in
            the schema. Defaults to 'STRING'.

        Returns
        -------
            A list of dictionaries containing field 'name' and 'type' entries, suitable for use in a
            BigQuery Tables resource schema.

        Notes
        -----
        Note that if you don't explicitly set
        the types of the columns in the dataframe, they may be of a type that forces coercion to
        STRING, so even though the fields in the dataframe themselves may be numeric, the type in the
        derived schema may not be. Hence it is prudent to make sure the Pandas dataframe is typed
        correctly.

        """

        type_mapping = {
            'i': 'INTEGER',
            'b': 'BOOLEAN',
            'f': 'FLOAT',
            'O': 'STRING',
            'S': 'STRING',
            'U': 'STRING',
            'M': 'TIMESTAMP'
        }

        fields = []
        for column_name, dtype in dataframe.dtypes.iteritems():
            fields.append({'name': column_name,
                           'type': type_mapping.get(dtype.kind, default_type)})

        return fields

    @staticmethod
    def from_dataframe(dataframe, default_type='STRING'):
        """
        Infer a BigQuery table schema from a Pandas dataframe. Note that if you don't explicitly set
        the types of the columns in the dataframe, they may be of a type that forces coercion to
        STRING, so even though the fields in the dataframe themselves may be numeric, the type in the
        derived schema may not be. Hence it is prudent to make sure the Pandas dataframe is typed
        correctly.

        Parameters
        ----------
        dataframe
            The DataFrame.
        default_type
            The default big query type in case the type of the column does not exist in
            the schema. Defaults to 'STRING'.

        Returns
        -------
        A :class:`~st_library.core.dataprovider.structured_data.schema.Schema` class itself.

        """
        return Schema(Schema._from_dataframe(dataframe, default_type=default_type))

    @staticmethod
    def _get_field_entry(name, value):
        entry = {'name': name}
        if isinstance(value, datetime.datetime):
            _type = 'TIMESTAMP'
        elif isinstance(value, bool):
            _type = 'BOOLEAN'
        elif isinstance(value, float):
            _type = 'FLOAT'
        elif isinstance(value, int):
            _type = 'INTEGER'
        elif isinstance(value, dict) or isinstance(value, list):
            _type = 'RECORD'
            entry['fields'] = Schema._from_record(value)
        else:
            _type = 'STRING'
        entry['type'] = _type
        return entry

    @staticmethod
    def _from_dict_record(data):
        """
        Infer a BigQuery table schema from a dictionary. If the dictionary has entries that
        are in turn OrderedDicts these will be turned into RECORD types. Ideally this will
        be an OrderedDict but it is not required.

        Parameters
        ----------
        data
            The dict to infer a schema from.

        Returns
        -------
        A list of dictionaries containing field 'name' and 'type' entries, suitable for use in a
        BigQuery Tables resource schema.

        """
        return [Schema._get_field_entry(name, value) for name, value in list(data.items())]

    @staticmethod
    def _from_list_record(data):
        """
        Infer a BigQuery table schema from a list of values.

        Parameters
        ----------
        data
            The list of values.

        Returns
        -------
        A list of dictionaries containing field 'name' and 'type' entries,
        suitable for use in a BigQuery Tables resource schema.

        """
        return [Schema._get_field_entry('Column%d' % (i + 1), value) for i, value in enumerate(data)]

    @staticmethod
    def _from_record(data):
        """
        Infer a BigQuery table schema from a list of fields or a dictionary. The typeof the elements
        is used. For a list, the field names are simply 'Column1', 'Column2', etc.

        Parameters
        ----------
        data
            The list of fields or dictionary.

        Returns
        -------
        A list of dictionaries containing field 'name' and 'type' entries, suitable for use in a
        BigQuery Tables resource schema.

        Raises
        ------
        exceptions.SchemaCreationError
            If a schema cannot be created

        """
        if isinstance(data, dict):
            return Schema._from_dict_record(data)
        elif isinstance(data, list):
            return Schema._from_list_record(data)
        else:
            raise exceptions.SchemaCreationError('Cannot create a schema from record %s' % str(data))

    @staticmethod
    def from_record(source):
        """
        Infers a table/view schema from a single record that can contain a list of fields or a
        dictionary of fields. The type of the elements is used for the types in the schema. For a
        dict, key names are used for column names while for a list, the field names are simply named
        'Column1', 'Column2', etc. Note that if using a dict you may want to use an OrderedDict
        to ensure column ordering is deterministic.

        Parameters
        ----------
        source
            The list of field values or dictionary of key/values.

        Returns
        -------
            A Schema for the data.

        """
        # TODO(gram): may want to allow an optional second argument which is a list of field
        # names; could be useful for the record-containing-list case.
        return Schema(Schema._from_record(source))

    @staticmethod
    def from_data(source):
        """
        Infers a table/view schema from its JSON representation, a list of records, or a Pandas dataframe.

        Parameters
        ----------
        source
            the Pandas Dataframe, a dictionary representing a record, a list of heterogeneous
            data (record) or homogeneous data (list of records) from which to infer the schema, or
            a definition of the schema as a list of dictionaries with 'name' and 'type' entries
            and possibly 'mode' and 'description' entries. Only used if no data argument was provided.
            'mode' can be 'NULLABLE', 'REQUIRED' or 'REPEATED'. For the allowed types, see:
            https://cloud.google.com/bigquery/preparing-data-for-bigquery#datatypes

        Returns
        -------
        A Schema for the data.

        Notes
        -----
        Note that there is potential ambiguity when passing a list of lists or a list of
        dicts between whether that should be treated as a list of records or a single record
        that is a list. The heuristic used is to check the length of the entries in the
        list; if they are equal then a list of records is assumed. To avoid this ambiguity
        you can instead use the Schema.from_record method which assumes a single record,
        in either list of values or dictionary of key-values form.

        Raises
        ------
        exceptions.SchemaCreationError
            If a schema cannot be created

        """
        if isinstance(source, pandas.DataFrame):
            bq_schema = Schema._from_dataframe(source)
        elif isinstance(source, list):
            if len(source) == 0:
                bq_schema = source
            elif all(isinstance(d, dict) for d in source):
                if all('name' in d and 'type' in d for d in source):
                    # It looks like a bq_schema; use it as-is.
                    bq_schema = source
                elif all(len(d) == len(source[0]) for d in source):
                    bq_schema = Schema._from_dict_record(source[0])
                else:
                    raise exceptions.SchemaCreationError(('Cannot create a schema from '
                                                          'heterogeneous list %s; perhaps you meant ' +
                                                          'to use Schema.from_record?') % str(source))
            elif isinstance(source[0], list) and \
                    all([isinstance(l, list) and len(l) == len(source[0]) for l in source]):
                # A list of lists all of the same length; treat first entry as a list record.
                bq_schema = Schema._from_record(source[0])
            else:
                # A heterogeneous list; treat as a record.
                raise exceptions.SchemaCreationError(('Cannot create a schema from heterogeneous '
                                                      'list %s; perhaps you meant ' +
                                                      'to use Schema.from_record?') % str(source))
        elif isinstance(source, dict):
            raise exceptions.SchemaCreationError(('Cannot create a schema from dict %s; perhaps you meant to use ' +
                                                  'Schema.from_record?') % str(source))
        else:
            raise exceptions.SchemaCreationError('Cannot create a schema from %s' % str(source))
        return Schema(bq_schema)

    def __getitem__(self, key):
        """Provides ability to lookup a schema field by position or by name.
        """
        if isinstance(key, basestring):
            return self._map.get(key, None)
        # noinspection PyCallByClass
        return list.__getitem__(self, key)

    def _add_field(self, name, data_type, mode='NULLABLE', description=''):
        field = Schema.Field(name, data_type, mode, description)
        self.append(field)
        self._map[name] = field

    def find(self, name):
        """
        Get the index of a field in the flattened list given its (fully-qualified) name.

        Parameters
        ----------
        name
            the fully-qualified name of the field.

        Returns
        -------
        The index of the field, if found; else -1.
        """
        for i in range(0, len(self)):
            if self[i].name == name:
                return i
        return -1

    def _populate_fields(self, data, prefix=''):
        for field_data in data:
            name = prefix + field_data['name']
            data_type = field_data['type']
            self._add_field(name, data_type, field_data.get('mode', None),
                            field_data.get('description', None))

            if data_type == 'RECORD':
                # Recurse into the nested fields, using this field's name as a prefix.
                self._populate_fields(field_data.get('fields'), name + '.')

    def __str__(self):
        """ Returns a string representation of the non-flattened form of the schema. """
        # TODO(gram): We should probably return the flattened form. There was a reason why this is
        # not but I don't remember what it was. Figure that out and fix this.
        return str(self._bq_schema)

    def __eq__(self, other):
        """ Compares two schema for equality. """
        other_map = other._map
        if len(self._map) != len(other_map):
            return False
        for name in self._map.keys():
            if name not in other_map:
                return False
            if not self._map[name] == other_map[name]:
                return False
        return True

    def __ne__(self, other):
        """ Compares two schema for inequality. """
        return not (self.__eq__(other))



import abc

import requests
import six
from requests.auth import HTTPBasicAuth
from urlobject import URLObject

from . import resources


@six.add_metaclass(abc.ABCMeta)
class BaseApiClient(object):
    GET = 'GET'
    POST = 'POST'

    def __init__(self, *args, **kwargs):
        self.logger = self._get_logger()

    @abc.abstractmethod
    def _get_logger(self):
        """
        Return a logger object

        """

    @abc.abstractmethod
    def _raise_http_error(self, *args, **kwargs):
        """
        Raise appropriate exception when HTTP error is occurred

        """

    @abc.abstractmethod
    def _raise_performing_request_error(self, *args, **kwargs):
        """
        Raise appropriate exception when Request cannot be performed

        """

    @abc.abstractmethod
    def _get_auth_string(self):
        """
        Return auth user string
        """

    def request(
            self,
            http_method,
            path,
            params=None,
            data=None,
            json=None,
            base=resources.METADATA_SERVICE_ENDPOINT,
            basic_auth_tuple=None,
            sec_id_for_special_token=None,
            extra_headers=None,
            raw_content=False
    ):
        assert not (basic_auth_tuple and sec_id_for_special_token), 'Only one of these args can be specified'

        if path.startswith('/'):
            path = path[1:]

        retry = False
        auth = None

        if basic_auth_tuple:
            # The basic auth can be used instead of normal Bearer auth
            auth = HTTPBasicAuth(*basic_auth_tuple)
            headers = {}
        else:
            # Normal Bearer auth
            if sec_id_for_special_token:
                assert hasattr(self, 'get_sec_access_token'), 'Method does not exist. Forgot Mixin?'
                # This param is overriding auth with SEC token auth
                headers = {'Authorization': 'Bearer {}'.format(self.get_sec_access_token(sec_id_for_special_token))}
                retry = True
            else:
                # Use user auth
                headers = {'Authorization': self._get_auth_string()}
                retry = False

        url = URLObject(base).add_path(path)
        headers['Content-Type'] = 'application/json'

        if extra_headers:
            headers.update(extra_headers)

        def _request():
            try:
                self.logger.info('Calling metadata service (%s) : %s', http_method, url)
                resp_ = requests.request(http_method, url,
                                         params=params, data=data, json=json, headers=headers, auth=auth)
            except requests.RequestException as e_:
                self.logger.info('Outer service request error: {}'.format(e_))
                self._raise_performing_request_error('Error of request performing.: {}'.format(e_))

            return resp_

        resp = _request()
        try:
            resp.raise_for_status()
        except requests.HTTPError as e:
            self.logger.info(resp.content)

            if retry and resp.status_code in [401, 403]:
                self._retry_occurred()
                # Probably sec_access_token is expired, let's get a new one
                headers = {'Authorization': 'Bearer {}'.format(self.get_sec_access_token(sec_id_for_special_token))}
                # Retry request
                resp = _request()
                try:
                    resp.raise_for_status()
                except requests.HTTPError as e:
                    # This is not access_token expiration, raise exception
                    self._raise_http_error(resp.status_code, resp.reason)
            else:
                self._raise_http_error(resp.status_code, resp.reason)

        except Exception as e:
            # This error should never be happened
            self.logger.info('Unexpected error: {}'.format(e))
            raise

        if not raw_content:
            return resp.json()

        return resp.content

    def _retry_occurred(self):
        """
        This method is overridden for test purposes. By default it's do nothing and don't need to be overridden.
        """



import json


class LibraryException(Exception):
    """
    Base exception class for the Library.

    """


class ServiceRequestError(LibraryException):
    """
    This exception is raised when request to the Service failed.

    """

    def __init__(self, status, content):
        self.status = status
        self.content = content
        self.message = u'HTTP request failed. Status code : {} '.format(status)
        # Try extract a message from the body; swallow possible resulting ValueErrors and KeyErrors.
        try:
            error = json.loads(content)['error']
            if 'errors' in error:
                error = error['errors'][0]
            self.message += u': ' + error['message']
        except Exception:
            if content:
                lines = content.split(u'\n')
                if lines:
                    self.message += u': ' + lines[0]

    def __str__(self):
        return self.message


class InternalError(LibraryException):
    """
    An internal logic error.

    """


class PerformanceRequired(LibraryException):
    """
    This exception is raised when Performance is mandatory for executing operation,
    but it is not defined for current Environment.

    """


class SchemaCreationError(LibraryException):
    """
    An error of schema creation

    """


class DatabaseConnectionError(LibraryException):
    """
    An error of connection to database

    """


class DatabaseNotFound(LibraryException):
    """
    An error raising after try to access the missing database
    """



import psycopg2

from st_library.utils.generics.connectors import ConnectorContainer


class Postgres(object):
    def __init__(self, name, host, port, username, password, uses_google_sql):
        self._name = name
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._uses_google_sql = uses_google_sql
        # TODO: uses_google_sql
        self._conn = psycopg2.connect(database=self._name,
                                      user=self._username,
                                      password=self._password,
                                      host=self._host, port=self._port)

    def __repr__(self):
        return '<Postgres db "{}">'.format(self._name)

    @property
    def name(self):
        return self._name

    @property
    def conn(self):
        return self._conn


class PostgresContainer(ConnectorContainer):
    def _do_initialize_data(self):
        assert not len(self._list)

        params = self._fetch_param_dict([
            'psql_host', 'psql_name', 'psql_username', 'psql_password', 'psql_port', 'uses_google_sql'
        ])

        list_of_servers_params = [params]

        for server in list_of_servers_params:
            obj = Postgres(server['psql_name'], server['psql_host'],
                           int(server['psql_port']), server['psql_username'],
                           server['psql_password'], self._get_bool_uses_google_sql(server['uses_google_sql']))
            self._list.append(obj)
            self._dict[obj.name] = obj

    def _get_bool_uses_google_sql(self, raw_uses_google_sql):
        if raw_uses_google_sql == '1':
            return True
        return False
